import React, {useState, useEffect, useMemo, useRef}  from 'react';
import logo from './logo.svg';
import './App.css';

import { GraphQLClient } from 'graphql-request';

import createModels2 from './front-models'

let gql = new GraphQLClient("/graphql", {headers: localStorage.authToken ? {Authorization: 'Bearer '+localStorage.authToken} : {}})


//TODO: use graphql-tag to get types, or, at least detect query 
//for proper binding between model objects and result of gql query



const dataReader = async () => {
    const { PriceItem } = await createModels2(gql)
    let data            = await PriceItem.find({}, {limit: [10]})
    return {data, PriceItem }
}

const Th = p =>
<div className='Th' {...p} />

const Row = ({top, children}) =>
<div className='Row' style={{position: 'relative',top}}>
    {children}
</div>

const Cell = ({children, ...props}) =>
<div className='Cell' {...props}>
    {children && children.toString()}
</div>


function ModelGrid({model, rowHeight, gridHeight, overload, query, cellDoubleClick}){
    const onScreenRowCount   = gridHeight/rowHeight
    const overloadedRowCount = overload * onScreenRowCount

    const [totalCount, setTotalCount] = useState(-1)


    const [gridContentRef, setGridContentRef] = useState()
    const [scroll, setScroll]                 = useState(0)



    const  onScreenFirstRowIndex = Math.floor(scroll/rowHeight)
    let skip                  =  onScreenFirstRowIndex  - (overloadedRowCount - onScreenRowCount)/2;

    if (skip < 0) skip = 0

    const [sort, setSort] = useState([])

    const [records, setRecords] = useState()
    const cursorCalls = {skip: [skip], limit: [overloadedRowCount]}



    if (sort.length) cursorCalls.sort = [sort]


    useEffect(() => {
        if (totalCount > 0) { 
            model.find(query, cursorCalls).then(records => setRecords(records))
        }
    },[model, query, sort, totalCount])

    useEffect(() => {
        model && model.count(query).then(count =>  setTotalCount(count))
    },[model, query, totalCount])

    if (!model) return <>No model</>;
    return(
        <>
            <div className='GridHeader'>
                {model.fields.map(field => <Th onClick={e => (setSort(field.name === sort[0] ? [sort[0], -sort[1]] : [field.name, 1]), setRecords(null))}>{field.name}</Th>)}
            </div>
            <div className='GridViewport' 
                 onScroll={e => {
                        if (Math.abs(e.target.scrollTop - scroll) > gridHeight*overload/3){  
                            setScroll(e.target.scrollTop); 
                            setRecords(null)
                        } }} 
                 style={{maxHeight: gridHeight, height: gridHeight}}>

                <div className='GridContent' 
                     style={{height: totalCount*rowHeight}}
                     ref={e => setGridContentRef(e)}>
                    {console.log(records), records && records.map((record,i) => 
                                                    <Row top={(skip)*rowHeight}>
                                                        {model.fields.map(field => 
                                                            <Cell onDoubleClick={e => cellDoubleClick(field.name, record[field.name], record._id)}>
                                                                {record[field.name] === 'object' ? record[field.name].toString() :  record[field.name]}
                                                            </Cell>)}
                                                    </Row>)}
                </div>
            </div>
        </>
    );
}

const ModelListItem = "div"

const ModelList = ({models, selected, onChange, Item=ModelListItem}) => {
    return (
        <aside>
            {Object.entries(models).map(([name, model]) => {
                return (
                    <Item key={name} 
                                   selected={name === selected}
                                   className={name === selected && "selected"}
                                   onClick = {() => onChange(name)}
                                   >
                        {name}
                    </Item>
                )
            })}
        </aside>
    )
}

const Search = ({...props}) => 
<input placeholder="Search..." {...props} className="Search"/>

const Count = ({...props}) => <div className="Count" {...props} /> 

const searchQueryBuilder = (search, model) => {
    const queryRegexp = `/${search.trim().split(/\s+/).join('|')}/`
    return {$or: model.fields.filter(field => field.type.kind == 'SCALAR').map(field => ({[field.name]: queryRegexp}))}
}

const GridHeaderItem = ({field, sort, ...props}) => 
<div className="GridHeaderItem" {...props}>
    {field.name in sort && sort[field.name] === -1 && '^ '}
    {field.name}
    {field.name in sort && sort[field.name] === 1 && ' v'}
</div>



const GridHeader = ({fields, sort, onSort}) => 
<div className="GridHeader">
    {fields.map(field => <GridHeaderItem field={field} sort={sort} onClick={() => onSort(field.name)}/>)}
</div>

const Grid = ({sort, scroll, gridHeight, overload, setScroll, count, rowHeight, records, skip, model, cellDoubleClick}) =>  {
    const viewport = useRef()
    return (
    <div className='GridViewport' 
         onScroll={e => {
             console.log((Math.abs(e.target.scrollTop - scroll)))
                if (Math.abs(e.target.scrollTop - scroll) > gridHeight*overload/5){  
                    setScroll(e.target.scrollTop); 
                    //setRecords(null)
                } }} 
         style={{maxHeight: gridHeight, height: gridHeight}}
         ref={viewport}>

        <div className='GridContent' 
             style={{height: count*rowHeight, minHeight: count*rowHeight, maxHeight: count*rowHeight}} >
            <div style={{height: skip*rowHeight}} />
            {records && records.map((record,i) => 
                                            <Row key={record._id || record.key}>
                                                {model.fields.map(field => 
                                                    <Cell onDoubleClick={e => cellDoubleClick(field.name, record[field.name], record._id)}>
                                                        {record[field.name] === 'object' ? record[field.name].toString() :  record[field.name]}
                                                    </Cell>)}
                                            </Row>)}
            <div style={{height: (count - skip - records.length)*rowHeight}} />
        </div>
    </div>
    )
}


const ModelView = ({model, components:Components={Search, Count, GridHeader, Grid}, rowHeight=50, gridHeight=500, overload=2}) => {
    const onScreenRowCount   = Math.ceil(gridHeight/rowHeight)
    const overloadedRowCount = Math.ceil(overload * onScreenRowCount)
    const [scroll, setScroll]                 = useState(0)



    const  onScreenFirstRowIndex = Math.floor(scroll/rowHeight)
    let skipAbove             = Math.floor((overloadedRowCount - onScreenRowCount)/2);
    let skip                  =  onScreenFirstRowIndex  - skipAbove 


    const [records, setRecords] = useState()


    const [search, setSearch] = useState("")
    const [count, setCount]   = useState(0)

    if (count - skip < overloadedRowCount) skip = count - overloadedRowCount 
    if (skip < 0) skip = 0

    const [query,  setQuery]  = useState({})
    const [cursorCalls,  setCursorCalls]  = useState({sort:[{}], skip: skip ? [skip] : undefined, limit: overloadedRowCount ? [overloadedRowCount] : undefined})

    console.log(cursorCalls)

    useEffect(() => {
        setCursorCalls({sort:cursorCalls.sort, skip: skip ? [skip] : undefined, limit: overloadedRowCount ? [overloadedRowCount] : undefined})
    }, [scroll])

    const timeout = useRef(0)
    useEffect(() => {
        clearInterval(timeout.current)
        timeout.current = setTimeout(() => {
            setQuery(searchQueryBuilder(search, model))
        }, 1000)
    },[search, model])

    useEffect(() => {
        model.count(query, cursorCalls).then(count => setCount(count))
        model.find(query, cursorCalls).then(records => Promise.all(records)).then(records => setRecords(records))
    }, [query, model, cursorCalls])

    //console.log(records)

    return (
        <>
                <Components.Search value={search} onChange={({target: {value}}) => setSearch(value)}/>
                <Components.Count> 
                    {count}
                </Components.Count> 
                <GridHeader fields={model.fields} 
                            sort={cursorCalls.sort} 
                            onSort={sort => setCursorCalls({...cursorCalls,
                                sort: {[sort]: cursorCalls.sort[sort] === 1 ? -1 : 1}
                            })}/>

                {records && <Grid 
                                scroll={scroll}
                                setScroll={setScroll}
                                gridHeight={gridHeight}
                                overload={overload}
                                count={count}
                                rowHeight={rowHeight}
                                records={records}
                                skip={skip}
                                model={model}
                                /> }
        </>
    )
}

const Admin = ({models, components:Components={ModelList, Search}}) => {
    const [selected, setSelected] = useState()


    return (
        <>
            <Components.ModelList models={models} onChange={(name) => setSelected(name)} selected={selected}/>
            <content>

                {selected && <ModelView model={models[selected]} /> }
            </content>
        </>
    )
}


function App() {
    let [models, setModels] = useState()
    let [queryText,  setQueryText]  = useState('{manufacturerName: "TOYOTA"}')
    let [query,  setQuery]  = useState({})
    models || createModels2(gql).then(models => setModels(models))

    console.log(query)
    const classes = models
    console.log(classes)
    //if (classes && Object.keys(classes).length){
        //classes.Good.find().then(goods => console.log(goods))
        //classes.Image.find().then(images => console.log(images))
    //}

    return (
        <div className="App">
            {models && <Admin models={models} />}
            {/*models && <ModelGrid model={models.Image} rowHeight={50} gridHeight={700} overload={5} query={query} 
                        cellDoubleClick={(name, text, _id) => setQueryText(`{${name}:\`${text}\`}`)}/> */}
            <textarea onChange={e => setQueryText(e.target.value)} value={queryText} />
            <button onClick={() => setQuery(eval(`(${queryText})`))}>Run</button>
        </div>
    );
}

export default App;
